#ifndef KBF_BUTTON_H
#define KBF_BUTTON_H

#include "kbf/gpio.h"
#include "kbf/rtos.h"

namespace kbf::driver {
    /**
     * @brief Button driver.
     */
    class Button {
    public:
        /** @brief Tag used for logging. */
        static const constexpr char *const TAG = "kbf::driver::Button";

        /**
         * @brief Default constructor.
         *
         * @param pin GPIO pin the button is connected to
         * @param longPressMs long press delay in milliseconds
         */
        explicit Button(int pin, int longPressMs = 1000);

        /** @brief GPIO pin in use. */
        const int pin;

        /** @brief Long press delay in milliseconds. */
        const int longPressMs;

        /**
         * @brief Called when button is pressed.
         *
         * @param button instance
         */
        void (*onPress)(Button &button) = nullptr;

        /**
         * @brief Called when button is released.
         *
         * @warning Will be called in parallel to onLongRelease() if it's defined after long button presses.
         *
         * @param button instance
         */
        void (*onRelease)(Button &button) = nullptr;

        /**
         * @brief Called when button is pressed and held for longPressDelayMs milliseconds.
         *
         * @param button instance
         */
        void (*onLongPress)(Button &button) = nullptr;

        /**
         * @brief Called when button is released after at least longPressDelayMs milliseconds.
         *
         * @warning Will be called in parallel to onRelease() if it's defined.
         *
         * @param button instance
         */
        void (*onLongRelease)(Button &button) = nullptr;

    private:
        kbf::gpio::Input input;

        static void handleHigh(kbf::gpio::Input &);

        static void handleLow(kbf::gpio::Input &);

        [[noreturn]] static void eventTaskHandler(void *);

        static const int      eventHigh = 0;
        static const int      eventLow  = 1;
        kbf::rtos::EventGroup eventGroup;  // TODO implement queue
        const kbf::rtos::Task *eventTask;

        static void longPressTaskHandler(void *);

        static const int      eventLongRelease = 0;
        kbf::rtos::EventGroup longPressEventGroup;
        kbf::rtos::Task       *longPressTask   = nullptr;
    };
}

#endif //KBF_BUTTON_H
