#ifndef KBF_LCD_H
#define KBF_LCD_H

#include <string>

#include <hd44780.h>

/**
 * @brief Device drivers.
 */
namespace kbf::driver {
    /**
     * @brief HD44780-compatible liquid crystal display driver.
     */
    class LCD {
    public:
        /** @brief Tag used for logging. */
        static constexpr const char *TAG = "kbf::driver::LCD";

        /**
         * @brief Constructor.
         *
         * @param rs RS pin
         * @param en E pin
         * @param d4 D4 pin
         * @param d5 D5 pin
         * @param d6 D6 pin
         * @param d7 D7 pin
         */
        LCD(uint8_t rs, uint8_t en, uint8_t d4, uint8_t d5, uint8_t d6, uint8_t d7);

        virtual ~LCD();

        /**
         * @brief Controls the display.
         *
         * @param on turn the display on/off
         * @param cursor turn underscore (_) cursor on/off
         * @param blink turn blinking cursor on/off
         */
        void control(bool on, bool cursor, bool blink);

        /**
         * @brief Clears the display and resets cursor to 0,0.
         */
        void clear();

        /**
         * @brief Moves the cursor.
         *
         * @param x column
         * @param y line
         */
        void move(int x, int y);

        /**
         * @brief Prints to the display.
         *
         * @tparam T data type
         * @param data data to print
         */
        template<typename T>
        void print(T data);

        /**
         * @brief Prints to the display.
         *
         * @tparam T data type
         * @param data data to print
         * @return this
         */
        template<typename T>
        LCD &operator<<(T data) {
            print(data);
            return *this;
        }

        /**
         * @brief Prints a string to a line, padding it with spaces at the end to erase any previously printed data.
         *
         * @param lineNumber
         */
        void println(const std::string&, int lineNumber);

        /**
         * @brief Scrolls the display content to the left by one character.
         */
        void scrollLeft();

        /**
         * @brief Scrolls the display content to the right by one character.
         */
        void scrollRight();

        /**
         * @brief Uploads a custom character to the display.
         *
         * @note Cursor is reset to 0,0 after upload.
         *
         * @param slot character slot; 0 <= n < 8
         * @param data character data
         */
        void setChr(int slot, const uint8_t data[8]);

        /**
         * @brief Helps printing custom characters.
         */
        class Chr {
        public:
            /**
             * @brief Returns the character corresponding to the given slot.
             * This can then be passed to print() or the stream operator.
             *
             * @param slot character slot; 0 <= n < 8
             * @return character
             */
            char operator[](int slot);
        };

        /**
         * @brief Helps printing custom characters.
         */
        Chr chr;

    private:
        hd44780_t lcd;
    };
}

#endif //KBF_LCD_H
