#ifndef KBF_HTTP_H
#define KBF_HTTP_H

#include <string>
#include <vector>
#include <map>
#include <memory>

#include <esp_http_server.h>
#include <nlohmann_json/json.hpp>


/**
 * @brief HTTP functions
 */
namespace kbf::http {
    /** @brief Tag used for logging. */
    static constexpr const char *TAG = "kbf::http";

    /**
     * @brief Supported HTTP methods.
     *
     * Mostly used for syntax consistency; values match that of httpd_method_t.
     */
    enum Method {
        GET     = 1,
        POST    = 3,
        OPTIONS = 6,
    };

    /**
     * @brief Base class for HTTP Request and Response objects.
     */
    class HTTPObject {
    public:
        /**
         * @brief HTTP headers
         *
         * @warning Parsing headers from the incoming response is not supported in kbf::http::Client (yet).
         *
         * @note Only the headers listed in acceptedHeaders will be parsed automatically for incoming
         * requests. For parsing custom headers, use readHeader().
         */
        std::map<std::string, std::string> headers;

        /** @brief body */
        std::string body;

        /**
         * @brief POST data
         */
        std::map<std::string, std::string> postData;

        /**
         * @brief Parses the body as a JSON object.
         *
         * @return json object or nullptr if the body is not a valid JSON object
         */
        [[nodiscard]] nlohmann::json json() const;

    protected:
        HTTPObject() = default;

        explicit HTTPObject(std::string body);

        static void parseToMap(std::map<std::string, std::string> &target, const std::string &buffer);
    };

    /**
     * @brief Handles a HTTP request.
     */
    class Request : public HTTPObject {
    public:
        /**
         * @brief Creates instance from a httpd_req_t object.
         *
         * @param httpdRequest
         */
        explicit Request(httpd_req_t *httpdRequest);

        /**
         * @brief Vector of headers automatically parsed for incoming requests.
         */
        static inline const std::vector<std::string> acceptedHeaders = {
                "Host",
                "User-Agent",
                "Content-Type",
        };

        /** @brief number of retry attempts in case of a socket read timeout */
        static inline const int MAX_READ_RETRY = 3;

        /**
         * @brief Parse a header from the httpd_req_t object. The key-value pair will also be added to the #headers map.
         *
         * @warning Can only be used on incoming requests.
         *
         * @param header
         * @return header std::string
         */
        std::string readHeader(const std::string &header);

        /** @brief HTTP method */
        Method method;

        /** @brief request URI */
        std::string uri;

        /** @brief query parameters */
        std::map<std::string, std::string> query{};


    private:
        httpd_req_t *httpdRequest;

        void readQuery();

        void readBody();
    };

    /**
     * @brief Stores a HTTP response.
     */
    class Response : public HTTPObject {
    public:
        /**
         * @brief Creates a Response from a char array.
         *
         * @param body response body; default is ""
         * @param status HTTP status code; default is 200
         * @param contentType value of Content-Type header; default is "text/html"
         */
        explicit Response(const char *body = "", int status = 200, std::string contentType = "text/html");

        /**
         * @brief Creates a Response from a string.
         *
         * @param body response body; default is ""
         * @param status HTTP status code; default is 200
         * @param contentType value of Content-Type header; default is "text/html"
         */
        explicit Response(std::string body, int status = 200, std::string contentType = "text/html");

        /**
         * @brief Creates a Response from a JSON object.
         *
         * @param json response body
         * @param status HTTP status code; default is 200
         * @param contentType value of Content-Type header; default is "application/json"
         */
        explicit Response(const nlohmann::json &json, int status = 200, std::string contentType = "application/json");

        /** @brief HTTP status code */
        int status; // TODO enumerate me

        /** @brief HTTP status text */
        std::string statusText = "OK";

        /** @brief value of Content-Type header */
        std::string contentType;
    };
}

#endif //KBF_HTTP_H
