#ifndef KBF_NET_H
#define KBF_NET_H

#include <string>
#include <vector>

#include <esp_netif.h>
#include <mdns.h>

/**
 * @brief General networking.
 */
namespace kbf::net {
    /** @brief Tag used for logging. */
    static constexpr const char *TAG = "kbf::net";

    /**
     * @brief Stores a MAC address.
     */
    class MAC {
    public:
        MAC() : addr{0, 0, 0, 0, 0, 0} {};

        explicit MAC(const uint8_t addr[6]);

        /** @brief MAC address as uint8_t[6] */
        uint8_t addr[6]{};

        /**
         * @brief Returns the MAC address as std::string.
         *
         * @return the address in "aa:bb:cc:dd:ee:ff" format
         */
        [[nodiscard]] std::string str() const;
    };

    /**
     * @brief Sets the base MAC address of the device.
     *
     * @note This function must be called before the WiFi driver is initialized,
     *       i.e. before kbf::wifi::init().
     *
     * @param MAC address to use
     */
    void setMac(const MAC &mac);

    /**
     * @brief Gets the base MAC address of the device.
     *
     * @return MAC address in use
     */
    MAC getMac();

    /**
     * @brief Stores an IPv4 address.
     */
    class IP {
    public:
        /** @brief Tag used for logging. */
        static const constexpr char *const TAG = "kbf::net::IP";

        IP() : addr{0} {};

        /**
         * @brief Instantiate from esp_ip4_addr.
         *
         * @param addr esp_ip4_addr
         */
        explicit IP(esp_ip4_addr addr) : valid(true), addr(addr.addr) {};

        /**
         * @brief Instantiate from string.
         *
         * @warning This constructor uses regex to parse the address, which requires a relatively large amount of
         * memory. This will lead to stack overflow if used in system events with a
         * default CONFIG_ESP_SYSTEM_EVENT_TASK_STACK_SIZE.
         *
         * @param address IP address in "123.123.123.123" format
         */
        explicit IP(const std::string &address);

        /**
         * @brief Returns the IPv4 address as a std::string.
         *
         * @return the address in "123.123.123.123" format
         */
        [[nodiscard]] std::string str() const;

        bool operator==(const IP &other) const { return addr == other.addr; }

        /**
         * @brief bool operator
         * @return true if IP address is valid
         */
        explicit operator bool() const { return valid; }

        /**
         * @brief not operator
         * @return true if IP address is invalid
         */
        bool operator!() const { return !valid; }

        /**
         * @brief Returns the nth byte of the IP address.
         *
         * @note Will return 0 if n < 0 or n > 3.
         *
         * @param n
         * @return nth byte
         */
        uint8_t operator[](const unsigned int n) const { return (addr & ((2 << 7) - 1) << n * 8) >> n * 8; }

    private:
        bool     valid = false;
        uint32_t addr;

        void invalidate(const std::string &);
    };

//    enum class Interface {
//        STA,
//        AP,
//        ETH
//    };
//
//    enum class IPVersion {
//        V4,
//        V6
//    };

    enum class Proto {
        TCP,
        UDP
    };

    /**
     * @brief Multicast DNS server.
     */
    class MDNS {
    public:
        /** @brief Tag used for logging. */
        static const constexpr char *const TAG = "kbf::net::MDNS";

        /** @brief mDNS server instance name. */
        static const constexpr char *const INSTANCE_NAME = "KBF mDNS server";

        /**
         * @brief Default constructor.
         *
         * Calls setHostname() if a hostname is passed.
         *
         * @param hostname hostname to set; default is ""
         */
        explicit MDNS(const std::string& hostname = "");

        /**
         * @brief Stops the mDNS server.
         */
        ~MDNS();

        /**
         * @brief Sets mDNS hostname.
         *
         * @param hostname
         */
        void setHostname(const std::string &hostname);

        /**
         * @brief Starts advertising a service.
         *
         * @param serviceType service type, e.g. "_http"
         * @param port
         * @param instanceName service instance name, e.g. "My HTTP Server"
         * @param proto
         */
        void addService(const std::string &serviceType, int port, const std::string &instanceName = "",
                        const Proto &proto = Proto::TCP
        );

        /**
         * @brief Queries A record for hostname.
         *
         * @param hostname
         * @param timeoutMs timeout in milliseconds; default is 2000
         * @return IP or nullptr if hostname not found
         */
        IP *queryA(const std::string &hostname, int timeoutMs = 2000);

        /**
         * @brief Holds a result of a PTR query.
         */
        struct PTRResult {
//            Interface       interface;
//            IPVersion       ipVersion;
            Proto           protocol;
            std::vector<IP> addresses;
            int             port;
            std::string     instanceName;
            std::string     hostName;
        };

        /**
         * @brief Queries PTR records for a service type.
         *
         * @param serviceType service type, e.g. "_http"
         * @param timeoutMs timeout in milliseconds
         * @param proto
         * @return vector of results
         */
        std::vector<PTRResult>
        queryPTR(const std::string &serviceType, int timeoutMs = 10000, Proto proto = Proto::TCP);

    private:
        static const char *protoStr(Proto proto);
    };
}

#endif //KBF_NET_H
