#ifndef KBF_NVS_H
#define KBF_NVS_H

#include <string>
#include <atomic>

#include <esp_log.h>
#include <nvs_flash.h>

namespace kbf::nvs {
    /**
     * @brief Non-Volatile Storage read / write functions.
     */
    class NVS {
    public:
        /**
         * @brief Constructor. Calls init() if it hasn't been called yet.
         *
         * @param name NVS namespace name; default is "storage"
         */
        explicit NVS(const std::string &name = "storage");

        ~NVS();

        /**
         * @brief Reads a value from NVS.
         *
         * @tparam T value type
         * @param key name
         * @param value value
         * @return true if successful; false if key not found
         */
        template<typename T>
        bool read(const std::string &key, T &value);

        /**
         * @brief Writes a value to NVS.
         *
         * @warning writing string values that contain '\0' doesn't work properly
         *
         * @tparam T value type
         * @param key name
         * @param value value
         */
        template<typename T>
        void write(const std::string &key, T &value);

        /** @brief NVS namespace name */
        const std::string name;

    private:
        nvs_handle_t handle{};
    };

    /**
     * @brief Initializes NVS. Will be called from NVS class ctor if it hasn't been called yet.
     */
    void init();

    /**
     * Erases NVS partition.
     *
     * @warning Erasing the NVS flash will invalidate all opened handles, i.e. all NVS class instances!
     */
    void erase();
}

#endif //KBF_NVS_H
