#ifndef KBF_WEB_SERVICE_H
#define KBF_WEB_SERVICE_H

#include <string>
#include <vector>

#include "kbf/http/server.h"
#include "kbf/macros.h"

namespace kbf {
    /**
     * @brief Controller-based HTTP server.
     */
    class WebService {
    public:
        /** @brief Tag used for logging. */
        static const constexpr char *TAG = "WebService";

        /**
         * @brief Constructor.
         *
         * @param port web server port; default is 80
         */
        explicit WebService(int port = 80);

        /**
         * @brief Destructor.
         *
         * Will call stop() if #running.
         */
        ~WebService();

        /**
         * @brief Starts the service.
         *
         * @note All Controllers must be attached before calling start().
         */
        void start();

        /**
         * @brief Stops the service.
         */
        void stop();

        class Controller;

        /**
         * @brief Attaches a Controller to the service.
         *
         * @tparam T Controller class
         */
        template<class T>
        void attach() {
            static_assert(std::is_base_of<Controller, T>::value, "attach(): type must be a Controller");

            Controller *controller = new T();

            void (*handler)(const http::Response &, void *) = nullptr;
            if ((void *) (controller->*(&Controller::onResponseSent)) !=
                (void *) (&Controller::onResponseSent)) { // NOLINT
                // set handler only if onResponseSent is implemented
                handler = Controller::responseSentHandler;
            }

            server.route({kbf::http::Method::GET, controller->path, handleGet, controller, handler});
            server.route({kbf::http::Method::POST, controller->path, handlePost, controller, handler});
        }

        /**
         * @brief Controller for the web service.
         *
         * Pass derived classes to WebService::attach().
         */
        class Controller {
            friend class WebService;

        protected:
            /**
             * @brief Constructor.
             *
             * Call from child constructor.
             *
             * @param path URI to handle
             */
            explicit Controller(std::string path);

            /** @brief URI to handle */
            const std::string path;

            /**
             * @brief Method to run on incoming GET requests.
             *
             * @param request
             * @return HTTP response
             */
            virtual kbf::http::Response get(const kbf::http::Request &request) {
                return kbf::http::Response("method not allowed", 405);
            }

            /**
             * @brief Method to run on incoming POST requests.
             *
             * @param request
             * @return HTTP response
             */
            virtual kbf::http::Response post(const kbf::http::Request &request) {
                return kbf::http::Response("method not allowed", 405);
            }

            /**
             * @brief Called after a response is sent successfully.
             *
             * @param response
             */
            virtual void onResponseSent(const kbf::http::Response &response) {}

        private:
            static void responseSentHandler(const kbf::http::Response &response, void *data);
        };

    private:
        kbf::http::Server server;

        static kbf::http::Response handleGet(const kbf::http::Request &request, void *data);

        static kbf::http::Response handlePost(const kbf::http::Request &request, void *data);

        bool running = false;
        int  port;
    };
}

#endif //KBF_WEB_SERVICE_H
