#include "kbf/wifi.h"

#include "kbf/macros.h"

using namespace kbf;

wifi::AP::AP(string ssid, string password, net::IP ip, net::IP netmask) :
        ssid(std::move(ssid)), password(std::move(password)), ip(ip), netmask(netmask) {
    ESP_LOGI(TAG, "AP()");

    std::copy(this->ssid.begin(), this->ssid.end() + 1, std::begin(config.ssid));
    std::copy(this->password.begin(), this->password.end() + 1, std::begin(config.password));

    config.ssid_len       = this->ssid.length();
    config.max_connection = 5; // TODO use Kconfig
    config.authmode       = WIFI_AUTH_WPA2_PSK; // TODO use AUTH_OPEN if no password is provided
}

wifi::AP::~AP() {
    ESP_LOGI(TAG, "~AP()");
    unregisterEventHandlers();
}

shared_ptr<wifi::AP>
wifi::AP::create(const string &ssid, const string &password, const net::IP &ip, const net::IP &netmask) {
    if (password.length() < 8) {
        ABORT("WiFi password needs to be at least 8 characters long.");
    }

    auto ap = new AP(ssid, password, ip, netmask);
    return shared_ptr<AP>(ap);
}

void wifi::AP::init() {
    netif = esp_netif_create_default_wifi_ap();

    setupDhcp();
    registerEventHandlers();
}

void wifi::AP::setupDhcp() {
    ESP_LOGD(TAG, "setupDhcp()");

    // TODO this should only be run once, before bringing up the interface for the first time; if the interface is already up, DHCP seems to fail silently

    esp_netif_ip_info_t ipInfo;
    IP4_ADDR(&ipInfo.ip, ip[0], ip[1], ip[2], ip[3]);
    IP4_ADDR(&ipInfo.gw, ip[0], ip[1], ip[2], ip[3]);
    IP4_ADDR(&ipInfo.netmask, netmask[0], netmask[1], netmask[2], netmask[3]);
    CHECK(esp_netif_dhcps_stop(netif));
    CHECK(esp_netif_set_ip_info(netif, &ipInfo));
    CHECK(esp_netif_dhcps_start(netif));
}

void wifi::AP::registerEventHandlers() {
    ESP_LOGD(TAG, "registerEventHandlers()");
    CHECK(esp_event_handler_instance_register(
            WIFI_EVENT, WIFI_EVENT_AP_START, &handleStart, this, &startHandler
    ));
    CHECK(esp_event_handler_instance_register(
            WIFI_EVENT, WIFI_EVENT_AP_STOP, &handleStop, this, &stopHandler
    ));
    CHECK(esp_event_handler_instance_register(
            WIFI_EVENT, WIFI_EVENT_AP_STACONNECTED, &handleConnect, this, &connectHandler
    ));
    CHECK(esp_event_handler_instance_register(
            WIFI_EVENT, WIFI_EVENT_AP_STADISCONNECTED, &handleDisconnect, this, &disconnectHandler
    ));
}

void wifi::AP::unregisterEventHandlers() {
    ESP_LOGD(TAG, "unregisterEventHandlers()");

    CHECK(esp_event_handler_instance_unregister(WIFI_EVENT, WIFI_EVENT_AP_START, &startHandler));
    CHECK(esp_event_handler_instance_unregister(WIFI_EVENT, WIFI_EVENT_AP_STOP, &stopHandler));
    CHECK(esp_event_handler_instance_unregister(WIFI_EVENT, WIFI_EVENT_AP_STACONNECTED, &connectHandler));
    CHECK(esp_event_handler_instance_unregister(WIFI_EVENT, WIFI_EVENT_AP_STADISCONNECTED, &disconnectHandler));

    // TODO why are these necessary? shouldn't instance unregistration be enough?
    CHECK(esp_event_handler_unregister(WIFI_EVENT, WIFI_EVENT_AP_START, &handleStart));
    CHECK(esp_event_handler_unregister(WIFI_EVENT, WIFI_EVENT_AP_STOP, &handleStop));
    CHECK(esp_event_handler_unregister(WIFI_EVENT, WIFI_EVENT_AP_STACONNECTED, &handleConnect));
    CHECK(esp_event_handler_unregister(WIFI_EVENT, WIFI_EVENT_AP_STADISCONNECTED, &handleDisconnect));
}

void wifi::AP::handleStart(void *arg, esp_event_base_t, int32_t, void *) {
    ESP_LOGD(TAG, "handleStart()");
    auto ap = static_cast<AP *>(arg);
    if (ap->onStart) ap->onStart();
}

void wifi::AP::handleStop(void *arg, esp_event_base_t, int32_t, void *) {
    ESP_LOGD(TAG, "handleStop()");
    auto ap = static_cast<AP *>(arg);
    if (ap->onStop) ap->onStop();
}

void wifi::AP::handleConnect(void *arg, esp_event_base_t, int32_t, void *eventData) {
    ESP_LOGD(TAG, "handleConnect()");
    auto ap  = static_cast<AP *>(arg);
    auto sta = STA(*static_cast<wifi_event_ap_staconnected_t *>(eventData));

    if (ap->stations.find(sta.aid) != ap->stations.end()) {
        ESP_LOGE(TAG, "sta.aid already in use: %d", sta.aid);
        ABORT("debug me");
    }

    if (ap->onConnect) ap->onConnect(sta);
    ap->stations.insert(std::pair<int, STA>(sta.aid, sta));
}

void wifi::AP::handleDisconnect(void *arg, esp_event_base_t, int32_t, void *eventData) {
    ESP_LOGD(TAG, "handleDisconnect()");
    auto ap  = static_cast<AP *>(arg);
    auto aid = (static_cast<wifi_event_ap_stadisconnected_t *>(eventData))->aid;

    if (ap->onDisconnect) ap->onDisconnect(ap->stations[aid]);
    ap->stations.erase(aid);
}

wifi::AP::STA::STA(wifi_event_ap_staconnected_t &eventData) : aid(eventData.aid), mac(eventData.mac) {}
